#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass
import numpy
from DataVisual import DataV


from PyMieSim.bin.Sets import CppCoreShellSet, CppCylinderSet, CppSphereSet, CppSourceSet, CppDetectorSet
from PyMieSim.bin.Experiment import CppExperiment
from PyMieSim.Tools.utils import ToList
from PyMieSim.LPmodes import LoadUnstructuredLPMode
from PyMieSim import Variables


@dataclass
class CoreShellSet():
    CoreDiameter: list
    """ Diameter of the core of the single scatterer [m]. """
    ShellDiameter: list
    """ Diameter of the shell of the single scatterer [m]. """
    CoreIndex: list = None
    """ Refractive index of the core of the scatterer. """
    ShellIndex: list = None
    """ Refractive index of the shell of the scatterer. """
    CoreMaterial: list = None
    """ Core material of which the scatterer is made of. Only if CoreIndex is not specified.  """
    ShellMaterial: list = None
    """ Shell material of which the scatterer is made of. Only if ShellIndex is not specified.  """
    nMedium: list = 1.0
    """ Refractive index of scatterer medium. """
    Name: str = 'CoreShell'
    """Name of the set """

    def __post_init__(self):
        self.BoundedCore = True if self.CoreMaterial is not None else False
        self.BoundedShell = True if self.ShellMaterial is not None else False


        assert numpy.max(self.CoreDiameter) < numpy.min(self.ShellDiameter), f'Core diameter [{numpy.max(self.CoreDiameter)}] cannot be superior to Shell diameter [{numpy.min(self.ShellDiameter)}]'

        self.kwargs = Variables.GetKwargs(CoreDiameter = self.CoreDiameter,
                                          ShellDiameter = self.ShellDiameter,
                                          CoreMaterial = self.CoreMaterial,
                                          ShellMaterial = self.ShellMaterial,
                                          CoreIndex = self.CoreIndex,
                                          ShellIndex = self.ShellIndex,
                                          nMedium = self.nMedium)



        if self.BoundedCore: del self.kwargs['CoreIndex']
        else:  del self.kwargs['CoreMaterial']

        if self.BoundedShell: del self.kwargs['ShellIndex']
        else:  del self.kwargs['ShellMaterial']

        self.Parameters = list(self.kwargs.values())


    def BindToExperiment(self, Experiment):
        Experiment.Binding.SetCoreShell( self.Binding )
        return self.Parameters


    def Evaluate(self, Wavelength):
        if self.BoundedCore is True:
            CoreIndex = numpy.asarray( [material.GetRI(Wavelength) for material in self.kwargs['CoreMaterial'] ] )

            self.kwargs['CoreMaterial'] = Variables.GetKwargs(CoreIndex = CoreIndex)['CoreIndex']

        if self.BoundedShell is True:
            ShellIndex = numpy.asarray( [material.GetRI(Wavelength) for material in self.kwargs['ShellMaterial'] ] )

            self.kwargs['ShellMaterial'] = Variables.GetKwargs(ShellIndex = ShellIndex)['ShellIndex']

        self.Binding = CppCoreShellSet(**{k: v.Values for k, v in self.kwargs.items()})



@dataclass
class SphereSet():
    Diameter: list
    """ Diameter of the single scatterer in unit of meter. """
    Index: list = None
    """ Refractive index of scatterer. """
    Material: list = None
    """ Material of which the scatterer is made of. Only if Index is not specified. """
    nMedium: list = 1.0
    """ Refractive index of scatterer medium. """
    Name: str = 'Sphere'
    """Name of the set """

    def __post_init__(self):

        self.BoundedIndex = True if self.Material is not None else False

        self.kwargs = Variables.GetKwargs(Diameter = self.Diameter,
                                          Material = self.Material,
                                          Index    = self.Index,
                                          nMedium  = self.nMedium)



        if self.BoundedIndex: del self.kwargs['Index']
        else:  del self.kwargs['Material']

        self.Parameters = list(self.kwargs.values())


    def BindToExperiment(self, Experiment):
        Experiment.Binding.SetSphere( self.Binding )


    def Evaluate(self, Wavelength):
        if self.BoundedIndex is True:
            Index = numpy.asarray( [material.GetRI(Wavelength) for material in self.kwargs['Material'] ] )

            self.kwargs['Material'] = Variables.GetKwargs(Index = Index)['Index']

        self.Binding = CppSphereSet(**{k: v.Values for k, v in self.kwargs.items()})




@dataclass
class CylinderSet():
    Diameter: list
    """ Diameter of the single scatterer in unit of meter. """
    Index: list = None
    """ Refractive index of scatterer. """
    Material: list = None
    """ Refractive index of scatterer medium. """
    nMedium: list = 1.0
    """ Material of which the scatterer is made of. Only if Index is not specified. """
    Name: str = 'Cylinder'
    """Name of the set """

    def __post_init__(self):

        self.BoundedIndex = True if self.Material is not None else False

        self.kwargs = Variables.GetKwargs(Diameter = self.Diameter,
                                          Material = self.Material,
                                          Index    = self.Index,
                                          nMedium  = self.nMedium)

        if self.BoundedIndex: del self.kwargs['Index']
        else:  del self.kwargs['Material']

        self.Parameters = list(self.kwargs.values())


    def BindToExperiment(self, Experiment):
        Experiment.Binding.SetCylinder( self.Binding )


    def Evaluate(self, Wavelength):
        if self.BoundedIndex is True:
            Index = numpy.asarray( [material.GetRI(Wavelength) for material in self.kwargs['Material'] ] )

            self.kwargs['Material'] = Variables.GetKwargs(Index = Index)['Index']

        self.Binding = CppCylinderSet(**{k: v.Values for k, v in self.kwargs.items()})





@dataclass
class SourceSet(object):
    Wavelength: float = 1.0
    """ Wavelenght of the light field. """
    Polarization: float = None
    """ Polarization of the light field in degree. """
    Amplitude: float = None
    """ Maximal value of the electric field at focus point. """
    Name: str = 'PlaneWave'
    """ Name of the set """

    def __post_init__(self):
        self.kwargs = Variables.GetKwargs(Wavelength = self.Wavelength,
                                          Polarization = self.Polarization,
                                          Amplitude    = self.Amplitude)

        self.Parameters = list(self.kwargs.values())
        Jones = numpy.array( [ (numpy.cos(numpy.deg2rad(x)), numpy.sin(numpy.deg2rad(x))) if x is not None else (numpy.nan, numpy.nan) for x in self.kwargs['Polarization'].Values ] )

        self.Binding = CppSourceSet(Wavelength = self.kwargs['Wavelength'].Values,
                                    Jones      = Jones.astype(complex),
                                    Amplitude  = self.kwargs['Amplitude'].Values)

    def BindToExperiment(self, Experiment):
        Experiment.Binding.SetSource( self.Binding )


@dataclass
class DetectorSet(object):
    ScalarField: numpy.ndarray
    """ Array representing the detection field distribution. """
    NA: list
    """ Numerical aperture of imaging system. """
    GammaOffset: list
    """ Angle [Degree] offset of detector in the direction perpendicular to polarization. """
    PhiOffset: list
    """ Angle [Degree] offset of detector in the direction parallel to polarization. """
    Filter: list
    """ Angle [Degree] of polarization filter in front of detector. """
    CouplingMode: str  = 'Point'
    """ Method for computing mode coupling. Either Point or Mean. """
    Coherent: bool = False
    """ Describe the detection scheme coherent or uncoherent. """
    Name: str  = "Generic detector"
    """ Name of the set """

    def __post_init__(self):
        self.Binding = CppDetectorSet(ScalarField   = self.ScalarField,
                                      NA            = self.kwargs['NA'].Values,
                                      PhiOffset     = numpy.deg2rad(self.kwargs['PhiOffset'].Values),
                                      GammaOffset   = numpy.deg2rad(self.kwargs['GammaOffset'].Values),
                                      Filter        = numpy.array( [ x if x is not None else numpy.nan for x in self.kwargs['Filter'].Values] ),
                                      PointCoupling = True if self.CouplingMode == 'Point' else False,
                                      Coherent      = self.Coherent)

        self.Parameters = list(self.kwargs.values())



    def BindToExperiment(self, Experiment):
        Experiment.Binding.SetDetector( self.Binding )



class PhotodiodeSet(DetectorSet):
    Name = "Photodiode"
    def __init__(self,
                 NA           : list,
                 PhiOffset    : list,
                 GammaOffset  : list,
                 Filter       : list,
                 Sampling     : list,
                 CouplingMode : str = 'Point',
                 Coherent     : bool = False):

        self.kwargs = Variables.GetKwargs(CField      = 'Photodiode',
                                          NA          = NA,
                                          PhiOffset   = PhiOffset,
                                          GammaOffset = GammaOffset,
                                          Filter      = Filter)

        super().__init__(ScalarField  = numpy.asarray( [ numpy.ones(Sampling) ] ),
                         NA           = NA,
                         PhiOffset    = PhiOffset,
                         GammaOffset  = GammaOffset,
                         Filter       = Filter,
                         CouplingMode = 'Point',
                         Coherent     = False)




class LPModeSet(DetectorSet):
    Name = "LPMode"

    def __init__(self,
                 Mode         : list,
                 NA           : list,
                 PhiOffset    : list,
                 GammaOffset  : list,
                 Filter       : list,
                 Sampling     : int,
                 Rotation     : float = 0,
                 CouplingMode : str   = 'Point'):

        Mode = ToList(Mode)

        ScalarFields = numpy.asarray( [self.LoadUnstructuredLPMode(mode, Sampling) for mode in Mode] )

        self.kwargs = Variables.GetKwargs(CField      = Mode,
                                          NA          = NA,
                                          PhiOffset   = PhiOffset,
                                          GammaOffset = GammaOffset,
                                          Filter      = Filter)

        super().__init__(ScalarField  = ScalarFields,
                         NA           = NA,
                         PhiOffset    = PhiOffset,
                         GammaOffset  = GammaOffset,
                         Filter       = Filter,
                         CouplingMode = CouplingMode,
                         Coherent     = True)






    def LoadUnstructuredLPMode(self, Mode: list, Sampling: int):

        return LoadUnstructuredLPMode(Mode, Sampling)


class Setup(object):
    def __init__(self, ScattererSet = None, SourceSet = None, DetectorSet = None):

        self.SourceSet    = SourceSet

        self.DetectorSet  = DetectorSet

        self.ScattererSet = ScattererSet

        self.ScattererSet.Evaluate(self.SourceSet.Wavelength)

        self.Binding = CppExperiment()

        self.BindElements()

        self.Xtable = self.SourceSet.Parameters + self.ScattererSet.Parameters

        if self.DetectorSet is not None:
            self.Xtable += self.DetectorSet.Parameters



    def BindElements(self):
        self.SourceSet.BindToExperiment(self)
        self.ScattererSet.BindToExperiment(self)

        if self.DetectorSet:
            self.DetectorSet.BindToExperiment(self)



    def Get(self, Input: list):
        """Methode generate array of the givens parameters as a function of
        all independent variables.

        Returns
        -------
        :class:`DataVisual.DataV`
            DataVisual array containing the structured data

        """

        Input = ToList(Input)

        self.Ytable = Input

        Array = []
        for prop in Input:
            prop = 'Get'+ self.ScattererSet.Name + prop.Name

            Array.append(getattr(self.Binding, prop)())

        Array = numpy.asarray(Array)
        Array = numpy.rollaxis(Array, 0, len(Array.shape) )


        for n, e in enumerate(self.Xtable):
            e.Position = n

        return DataV(Array, Xtable=self.Xtable, Ytable=self.Ytable, WaterMark='PyMieSim')
