import numpy as np
from scipy import ndimage
import re
from inspect import signature

import PyMieSim
from PyMieSim.Physics           import Angle


class ClassRepr(object):

    Description  = ''
    ReprNoEval   = []
    ReprEval     = []
    HLine        = "\n" + "--"*20 + "\n"
    Methods      = []


    def __str__(self):
        return self.__repr__()


    def __repr__(self):
        self.ReprEval = [f"{item:12s}" + f": \t{getattr(self, item)}" for item in self.ReprEval]

        self.Methods = [e + f"{signature(getattr(self, e))}" for e in self.Methods]
        
        String = []
        String.append(f"{self.Description}")
        String.append(self.HLine)
        String.append("Attributes:\n\t")
        String.append( f"\n\t".join(self.ReprEval) + '\n\t' )
        String.append( f"\n\t".join(self.ReprNoEval) )
        String.append("\n\nMethods:\n\t")
        String.append( f"\n\t".join(self.Methods) )
        String.append(self.HLine)
        return "".join(String)


def Deg2Rad(Value):
    if Value is None: 
        return np.nan
    else: 
        return np.deg2rad(Value)

def Norm(Scalar):
    return np.sqrt(np.sum(np.abs(Scalar)**2))


def Normalize(Scalar):
    norm = Norm(Scalar)
    if norm == 0 or np.isnan(norm):
        return [0,0,0]
    else:
        return Scalar / norm


def InterpFull(Meshes, Scalar, Shape):

    Phi, Theta = np.mgrid[-np.pi/2:np.pi/2:complex(Shape[0]),
                          -np.pi:np.pi:complex(Shape[1])]

    Scalar = interp_at(Meshes.Phi.Radian,
                       Meshes.Theta.Radian,
                       Scalar.astype(np.complex).flatten(),
                       Phi.flatten(),
                       Theta.flatten(),
                       algorithm='linear',
                       extrapolate=True)

    return Scalar.reshape(Shape), Phi, Theta


def RescaleComplex(Input, Num):
    Scale = Num/Input.shape[0]
    InputReal = ndimage.interpolation.zoom(input=Input.real, zoom=(Scale), order = 2)
    InputImag = ndimage.interpolation.zoom(input=Input.imag, zoom=(Scale), order = 2)
    return InputReal + 1j * InputImag


def RotateComplex(Input, Rotation):
    InputReal =  ndimage.rotate(Input.real, Rotation, reshape=False)
    InputImag =  ndimage.rotate(Input.imag, Rotation, reshape=False)
    return InputReal + 1j * InputImag


def Angle2Direct(AngleVec: np.ndarray, k: float,) -> np.ndarray:

    RadSpace = np.deg2rad(AngleVec)

    FourierSpace = np.sin(RadSpace) * k / (2 * np.pi)

    fourier_unit = (FourierSpace[1] - FourierSpace[0]).__abs__()

    DirectSpace = np.fft.fftshift( np.fft.fftfreq( AngleVec.shape[0], d = fourier_unit ) )

    return DirectSpace



def Direct2Angle(DirectVec: np.ndarray, k: float) -> np.ndarray:

    direct_unit = (DirectVec[1] - DirectVec[0]).__abs__()

    FourierSpace = np.fft.fftshift( np.fft.fftfreq( DirectVec.shape[0], d = direct_unit ) )

    AngleVec = np.arcsin(2 * np.pi * FourierSpace / k) # conversion spatial frequency to angular space

    if np.isnan(AngleVec).any():
        raise Exception("Magnification too large.")

    return AngleVec * 180 / np.pi


def NA2Angle(NA: float) -> np.ndarray:
    if NA <=1.0: return Angle( np.arcsin(NA), Unit='Radian')
    if NA >= 1.0: return Angle( np.arcsin(NA-1) + np.pi/2, Unit='Radian')


def Direct2spherical(X, Y, MaxAngle):
    Z = 50 / np.tan(MaxAngle)

    _, Phi, Theta = Cart2Sp(X, Y, X*0+Z)

    return Phi, Theta

def Direct2Angle(X, Y, MaxAngle):
    MaxZ = np.max(X) / np.cos(MaxAngle)



def AngleUnit2DirectUnit(Angle, k):
    FourierSpace = np.sin(Angle) * k / (2 * np.pi)

    fourier_unit = (FourierSpace[1] - FourierSpace[0]).__abs__()

    DirectSpace = np.fft.fftshift( np.fft.fftfreq( Angle.shape[0], d = fourier_unit ) )

    return DirectSpace


def Cart2Sp(x, y, z):
    R = np.sqrt(x**2+y**2+z**2)
    Phi = np.arcsin(z/R)
    Theta = np.arctan2(y, x)
    return R, Phi, Theta


def Sp2Cart(Phi, Theta, R=None):
    R = R if R is not None else Phi*0+1
    x = R*np.cos(Phi)*np.cos(Theta)
    y = R*np.cos(Phi)*np.sin(Theta)
    z = R*np.sin(Phi)
    return x, y, z


def RotateY(Phi, Theta, Angle):
    x, y, z = Sp2Cart(Phi=Phi, Theta=Theta)
    xp = x*np.cos(Angle) + z*np.sin(Angle)
    yp = y
    zp = z*np.cos(Angle) - x*np.sin(Angle)
    return Cart2Sp(x=xp, y=yp, z=zp)


def RotateZ(Phi, Theta, Angle):
    x, y, z = Sp2Cart(Phi=Phi, Theta=Theta)
    xp = x*np.cos(Angle) - y*np.sin(Angle);
    yp = x*np.sin(Angle) + y*np.cos(Angle);
    zp = z;
    return Cart2Sp(x=xp, y=yp, z=zp)


def RotateX(Phi, Theta, Angle):
    x, y, z = Sp2Cart(Phi=Phi, Theta=Theta)

    xp = x;
    yp = y*np.cos(Angle) - z*np.sin(Angle);
    zp = y*np.sin(Angle) + z*np.cos(Angle);
    return Cart2Sp(x=xp, y=yp, z=zp)



def ToList(arg):
    if isinstance(arg, np.ndarray) and arg.ndim != 0:
        return arg

    if isinstance(arg, list):
        return np.asarray(arg)


    return np.asarray([arg])


def FormatStr(function):
    def wrapped(*args, **kwargs):
        args = (re.sub(r"\s+", "", arg.lower() ) if isinstance(arg, str) else arg for arg in args)

        kwargs = {k: re.sub(r"\s+", "", v.lower() ) if isinstance(v, str) else v for k, v in kwargs.items()}

        return function(*args, **kwargs)
    return wrapped


def GetSphericalMesh(Sampling, MaxAngle):

    x, y = np.mgrid[-50: 50: complex(Sampling), -50: 50: complex(Sampling)]
    z = 50 / np.tan(MaxAngle)
    _, theta, phi = Cart2Sp(x, y, x*0+z)

    return phi, theta


def Angle2Jones(Delta):
    val = exp(1j*Delta)*2
    JonesVector = np.array([1, val])
    Norm = (np.sqrt(1 + np.abs(val)**2 ))
    return JonesVector/Norm




# -
